import random
import string

from silica_django.mixins import JsonSchemaMixin


class SilicaUiElementType:
    horizontal = "HorizontalLayout"
    vertical = "VerticalLayout"
    group = "Group"
    category = "Category"
    categorization = "Categorization"
    control = "Control"
    custom_element = "CustomHTMLElement"


class SilicaUiElement:
    type = None
    kwargs = None
    field_name = None

    def __init__(self, *args, **kwargs):
        if self.kwargs is None:
            self.kwargs = {
                **kwargs
            }
            
    def get_ui_schema(self, silica_form):
        raise NotImplemented


class Control(SilicaUiElement, JsonSchemaMixin):
    type = SilicaUiElementType.control

    def __init__(self, field_name, scope=None, **kwargs):
        self.field_name = field_name
        super().__init__(**kwargs)
        if scope is None:
            scope = f'#/properties/{field_name.lower()}'
        self.kwargs.update({
            'scope': scope,
            'type': self.type,
            'field_name': field_name,
        })
        self.kwargs.update(kwargs)

    def get_ui_schema(self, silica_form):
        field_config = silica_form.get_field_config(self.field_name)
        if field_config:
            if field_config.rule:
                self.kwargs['rule'] = field_config.rule.get_rule_schema()
            self.kwargs.update(self._django_widget_to_ui_schema(silica_form.fields[self.field_name], field_config=field_config))
        else:
            self.kwargs.update(self._django_widget_to_ui_schema(silica_form.fields[self.field_name]))
        return self.kwargs


class SilicaLayout(SilicaUiElement):
    elements = None
    # because layouts are not named and therefore do not have a SilicaFieldConfig, css_classes and rule must be manually
    # set
    rule = None
    css_classes = None

    def __init__(self, *args, rule=None, css_classes=None, **kwargs):
        super().__init__(**kwargs)
        self.rule = rule
        self.css_classes = css_classes
        # args should be a list of SilicaUiElements
        self.elements = [self._process_arg(a) for a in args]
        self.kwargs.update({'type': self.type})

    @staticmethod
    def _process_arg(arg):
        # if arg is a SilicaUiElement, we are recursing through a layout; if it is a string, we have to construct a Control
        if type(arg) == str:
            return Control(arg)
        elif isinstance(arg, SilicaUiElement):
            return arg
        else:
            raise Exception(f"Unhandled type {type(arg)}")

    def get_all_elements(self):
        """ Returns all LayoutElements in a flat array; for use when nesting is not important, e.g. setting up mappings """
        elems = []
        for item in self.elements:
            if isinstance(item, SilicaLayout):
                elems.extend(item.get_all_elements())
            elif isinstance(item, SilicaUiElement):
                elems.append(item)
            else:
                raise Exception(f"Unsupported element {item}")
        return elems

    def get_ui_schema(self, silica_form):
        schema = self.kwargs
        # flatten elements
        schema['elements'] = [element.get_ui_schema(silica_form) for element in self.elements]
        if self.css_classes:
            schema['options'] = {'overrideCss': self.css_classes}
        if self.rule:
            schema['rule'] = self.rule.get_rule_schema()
        return schema


class HorizontalLayout(SilicaLayout):
    type = SilicaUiElementType.horizontal


class VerticalLayout(SilicaLayout):
    type = SilicaUiElementType.vertical


class Group(SilicaLayout):
    type = SilicaUiElementType.group

    def __init__(self, label, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.kwargs.update({'label': label})


class Categorization(SilicaLayout):
    type = SilicaUiElementType.categorization

    def get_ui_schema(self, silica_form):
        if any([e.type != SilicaUiElementType.category for e in self.elements]):
            raise Exception("Categorization elements may not have any non-Category direct children.")
        return super().get_ui_schema(silica_form)


class Category(SilicaLayout):
    type = SilicaUiElementType.category
    label = None

    def __init__(self, label, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.kwargs.update({'label': label})


class CustomHTMLElement(SilicaUiElement):
    type = SilicaUiElementType.custom_element
    content = None
    rule = None
    _id = None

    def __init__(self, content, *args, **kwargs):
        super().__init__(*args, **kwargs)
        letters = string.ascii_lowercase
        self._id = ''.join(random.choice(letters) for _ in range(10))
        self.kwargs.update({
            'name': self._id,
            'type': self.type,
            'scope': '#/'
        })
        self.kwargs.update(kwargs)
        self.content = content

    def get_mapped_content(self):
        """ returns a dictionary of this element's generated id to the content it should render """
        return {self._id: self.content}

    def get_ui_schema(self, silica_form):
        schema = self.kwargs
        if self.rule:
            schema['rule'] = self.rule.get_rule_schema()
        return schema



from silica_django.utils.jsonschema import JsonSchemaUtils


class UIEffects:
    show = "SHOW"
    hide = "HIDE"
    disable = "DISABLE"
    enable = "ENABLE"


class Condition(JsonSchemaUtils):
    """
    Conditions take the following form:
        Condition(key1=value1, key2=value2, ...)
    Examples:
        And(key1=1, key2='and') -> evaluates to True if key1 == 1 and key2 == 'and'
        Or(key1=1, key2=2) -> evaluates to True if key1 == 1 or key2 == 2
    Conditions are also composable, e.g.
        Or(And(key1=1, key2=1), key2=3) -> evaluates to True if (key1 == 1 and key2 == 1) or key2 == 3

    The JsonSchema Condition spec is here: https://json-schema.org/understanding-json-schema/reference/combining.html

    """

    args = None
    schema_key = None
    kwargs = None

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def get_condition_schema(self):
        schema = []
        if self.args:
            schema = self._process_args()
        if self.kwargs:
            schema.append(self._process_kwargs())
        return {
            self.schema_key: schema
        }

    def _process_kwargs(self):
        # kwargs are direct assignments of keys to values, so we can just return a dictionary
        return {
            "type": "object",
            "properties": {key: self.value_as_jsonschema(val) for key, val in self.kwargs.items()}
        }

    def _process_args(self):
        # args can only be other conditions
        return [
            arg.get_condition_schema() for arg in self.args
        ]


class Or(Condition):
    schema_key = "anyOf"


class And(Condition):
    schema_key = "allOf"


class Not(Condition):
    schema_key = "not"

    def get_condition_schema(self):
        # the Not JsonSchema object has to be specially formatted
        schema = {}
        if self.kwargs:
            schema.update({
                **self._process_kwargs(),
            })
        for arg in self.args:
            schema.update(**arg.get_condition_schema())
        return {
            self.schema_key: schema
        }


class Rule(JsonSchemaUtils):
    """
        Rules take in any number of conditions as args; any kwargs are treated as a single And, i.e. And(**kwargs)

        In order to support this behavior, all conditions are wrapped in an Or.
    """
    args = None
    effect = None
    custom_schema = None
    kwargs = None

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def get_rule_schema(self):
        schema_args = []
        if len(self.args):
            schema_args += list(self.args)
        if len(self.kwargs):
            schema_args.append(And(**self.kwargs))
        schema = Or(*schema_args).get_condition_schema()
        return {
            "effect": self.effect,
            "condition": {
                "scope": "#",
                "schema": schema
            }
        }


class ShowIf(Rule):
    effect = UIEffects.show


class HideIf(Rule):
    effect = UIEffects.hide


class DisableIf(Rule):
    effect = UIEffects.disable


class EnableIf(Rule):
    effect = UIEffects.enable



from collections import defaultdict

from django import forms
from django.core.exceptions import ValidationError
from django.db import transaction


class SilicaSubFormArrayField(forms.Field):
    """
        Implements a special kind of array field where the items are objects related to the form instance.

        This field does not behave like standard form fields, because its behavior is fundamentally different
        from that of a standard form field, which acts on a single instance.

        Array saves are done in a single transaction. Any object which already has a primary key field gets updated; 
        any object without a primary key field gets created. Any object in the queryset whose value is not present in
        the data gets deleted (this may change).

        To customize the behavior of this field, subclass it and implement your own handler functions as needed.

        TODO: support usage in non-model Forms

    """

    instance_form = None
    identifier_field = 'pk'
    queryset = None
    batch_size = 200
    min_instances = 0
    max_instances = None
    # this value must be set by the initialization of the form
    parent_instance = None

    _instantiated_forms = []
    # the list of update errors by pk of object
    _update_errors = defaultdict(list)
    # the list of errors for this field (database errors)
    _errors = []
    _qs_lookup = None
    _raw = None

    def __init__(self, *args, queryset=None, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.instance_form:
            raise NotImplementedError("You must define instance_form to use this field")
        if not issubclass(self.instance_form, forms.models.ModelForm):
            raise TypeError("instance_form must subclass ModelForm")
        if queryset:
            self.queryset = queryset

    def get_queryset(self):
        if self.queryset:
            # this should force a refresh of the queryset
            return self.queryset.all()
        else:
            return self.instance_form.Meta.model.objects.all()

    @property
    def qs_lookup(self):
        if not self._qs_lookup:
            self._qs_lookup = {item[self.identifier_field]: item for item in self.get_queryset()}
        return self._qs_lookup

    def prepare_for_commit(self, data):
        updates = []
        creates = []
        if data:
            for item in data:
                # the identifier field will either be the empty string or the correct value for the object
                pk = item.pop(self.identifier_field, None)
                if pk:
                    # if the item already has a pk, we are updating
                    update = self.handle_update(pk, item)
                    if update:
                        updates.append(update)
                else:
                    # if the item does not have a pk, we are creating
                    create = self.handle_create(item)
                    if create:
                        creates.append(create)
        return creates, updates

    def handle_delete(self, data):
        items_to_delete = self.get_items_to_delete(data)
        self.perform_delete(items_to_delete)

    def get_items_to_delete(self, data):
        if not data:
            data = []
        key = self.identifier_field
        item_keys = [datum[key] for datum in data if key in datum]
        return self.queryset.exclude(**{f'{key}__in': item_keys})

    def perform_delete(self, qs):
        """
        Actually performs the deletion operation on the queryset returned by handle_delete

        Args:
            qs: the queryset to be deleted
        """
        # add flag to turn off atomic operations (what to do for partial failures)
        # CRUD operations should be atomic with parent form
        try:
            qs.delete()
        except Exception as e:
            self._errors.append(f"There was an error deleting items. {repr(e)}")

    def prepare_create(self, item):
        """

        Args:
            item: a dictionary representing the keys and values from the submitted form

        Returns:
            A Django Model object which can be passed to bulk_create
        """
        return self.queryset.model(item)

    def handle_create(self, item):
        form = self.validate_against_form(item)
        if not form.errors:
            return self.prepare_create(form.cleaned_data)
        else:
            self._errors.append(f"There was an error creating an item. {form.errors}")
            return None

    def prepare_update(self, pk, item):
        """

        Args:
            pk: the primary key of the object to be updated
            item: a dictionary representing the keys and values from the submitted form

        Returns:
            A Django Model object which can be passed to bulk_update
        """
        return self.queryset.model(item, **{self.identifier_field: pk})

    def handle_update(self, pk, item):
        instance = self.qs_lookup[pk]
        form = self.validate_against_form(item, instance=instance)
        if not form.errors:
            return self.prepare_update(pk, form.cleaned_data)
        else:
            self._update_errors[pk].append(form.errors)
            return None

    def _instantiate_form(self, data=None, instance=None):
        # local import required to prevent cyclical imports
        from silica_django.forms import SilicaFormMixin
        # if the form subclasses silica form, include parent_instance as a kwarg
        kwargs = {'data': data, 'instance': instance}
        if issubclass(self.instance_form, SilicaFormMixin):
            kwargs['parent_instance'] = self.parent_instance
        return self.instance_form(**kwargs)

    def refresh_data(self):
        self.queryset = self.get_queryset()
        self._instantiated_forms = [self._instantiate_form(instance=instance) for instance in self.queryset]
        self.initial = [{**form.initial, f'{self.identifier_field}': getattr(form.instance, self.identifier_field)}
                        for form in self._instantiated_forms]

    def validate_against_form(self, form_data, instance=None):
        """ Ensure that data being passed from frontend validates against form """
        form = self._instantiate_form(data=form_data, instance=instance)
        form.is_valid()
        return form

    def validate(self, value):
        # runs after to_python, takes a value after coercion and raises validate error on any error, does not return anything
        super().validate(value)
        if value:
            errors = []
            for form in value:
                if form.errors:
                    errors.append(ValidationError(form.errors))
            if self._errors:
                errors.append(ValidationError([ValidationError(err) for err in self._errors]))
            if errors:
                raise ValidationError(errors)

    def data_as_forms(self, data):
        # transforms raw data into a list of forms
        forms = []
        for item in data:
            # the identifier field will either be the empty string or the correct value for the object
            pk = item.pop(self.identifier_field, None)
            instance = None
            if pk:
                instance = self.qs_lookup[pk]
            form = self.validate_against_form(item, instance=instance)
            forms.append(form)
        return forms

    def do_save(self):
        data = self._raw
        self.refresh_data()
        with transaction.atomic():
            # handle deletes
            self.handle_delete(data)
            creates, updates = self.prepare_for_commit(data)
            if creates:
                try:
                    self.queryset.model.objects.bulk_create(creates, batch_size=self.batch_size)
                except Exception as e:
                    self._errors.append(f"There was an error creating new objects. {repr(e)}")
            if updates:
                try:
                    self.queryset.model.objects.bulk_update(updates, self.instance_form._meta.fields,
                                                            batch_size=self.batch_size)
                except Exception as e:
                    self._errors.append(f"There was an error updating existing objects. {repr(e)}")
        # force form to re-do queryset and re-calculate initial values so that newly created & updated data is displayed on refresh
        self.refresh_data()

    def to_python(self, data):
        self._raw = data
        if data in self.empty_values:
            return None
        return self.data_as_forms(data)



class SilicaConfig:
    """
        A SilicaConfig object contains all the Silica-specific configuration options for the form as a whole.
        For now, no configuration is available, but planned features include the ability to set the JSONForms
        validation mode.
    """
    
    config = None
    
    def __init__(self, **kwargs):
        if self.config is None:
            self.config = {}
        # key is field_name, value is SilicaFieldConfig for that field
        self.config.update(kwargs)

    def get_field_config(self, field_name):
        return self.config.get(field_name, None)


class SilicaFieldConfig:
    """
    
        A SilicaFieldConfig object contains all the Silica-specific configuration options
        for a control. Note that this is used to update and override the generated
        schema, but all fields are optional; a silica-powered form will always generate
        enough for a functional render on its own.
    """
    
    rule = None
    schema = None
    uischema = None
    
    def __init__(self, rule=None, maximum=None, minimum=None, default=None, min_length=None, 
                 max_length=None, description=None, type=None, schema_format=None, label=None, 
                 scope=None, ui_options=None, detail=None, show_sort_buttons=None, element_label_prop=None, 
                 ui_format=None, readonly=None, multiple_of=None, title=None, examples=None, display_delete=None, 
                 enable_add=None, error_message=None, no_data_msg=None, static_title=None, add_text=None, 
                 max_item_text=None, css_classes=None, wrapper_css_classes=None):
        # build kwargs into the uischema and schema objects formatted as jsonschema expects
        self.rule = rule
        schema = {
            'maximum': maximum,
            'minimum': minimum,
            'default': default,
            'minLength': min_length,
            'maxLength': max_length,
            'description': description,
            'type': type,
            'format': schema_format,
            'multipleOf': multiple_of,
            'examples': examples,
            'title': title,
            'errorMessage': error_message,
        }
        uischema_options = ui_options
        uischema = {
            'label': label,
            'scope': scope,
            'options': {
                'detail': detail,
                'showSortButtons': show_sort_buttons,
                'elementLabelProp': element_label_prop,
                'format': ui_format,
                # silica custom UISchema properties
                'readOnly': readonly,
                'displayDelete': display_delete,
                'enableAddButton': enable_add,
                'noDataMsg': no_data_msg,
                'staticTitle': static_title,
                'addText': add_text,
                'maxItemText': max_item_text,
                'overrideCss': css_classes,
                'wrapperOverrideCss': wrapper_css_classes,
            }
        }
        if uischema_options:
            uischema['options'].update(uischema_options)
        self.schema = {k: v for k, v in schema.items() if v is not None}
        uischema['options'] = {k: v for k, v in uischema['options'].items() if v is not None}
        self.uischema = {k: v for k, v in uischema.items() if v is not None}



import django.forms


class SilicaRenderer(django.forms.Widget):
    custom_component_name = None

    def __init__(self, *args, custom_component_name=None, **kwargs):
        super().__init__(*args, **kwargs)
        if custom_component_name:
            self.custom_component_name = custom_component_name


class SilicaSubmitRenderer(SilicaRenderer, django.forms.CharField):
    custom_component_name = "silica-submit-renderer"



from collections import defaultdict

from django import forms

from silica_django.fields import SilicaSubFormArrayField
from silica_django.layout import Control, VerticalLayout, CustomHTMLElement
from silica_django.mixins import JsonSchemaMixin


class SilicaFormMixin(JsonSchemaMixin, forms.Form):
    """ Adds Silica functionality to any Django form.

        Required Meta fields:
        @fields - a list of fields you wish to be rendered. These should correspond to the form's @fields attribute.

        Optional Meta fields:
        @rules - a mapping of fields to the rules which should be applied to controls.
        @layout - a custom SilicaLayout. This will be used instead of generating a layout from your fields. 
                         Note that rules will still be applied.
        @custom_ui_schema - a mapping of fields to a dictionary matching the UISchema pattern.
        @custom_components - a mapping of fields to the name of the custom Control renderer you want to use.

     """
    class Meta:
        silica_config = None
        layout = None

    def __init__(self, *args, parent_instance=None, **kwargs):
        # if this form is an array item, it should have access to the instance of the form containing the array field
        self.parent_instance = parent_instance
        # if we are intaking data from a POST via args, process it
        new_args = [*args]
        new_kwargs = kwargs.copy()
        if len(args) > 0:
            # we have to mutate the querydict, so make a copy
            raw_data = new_args[0].copy()
            array_keys, array_info = self._extract_array_info(raw_data)
            for key, values in array_info.items():
                raw_data[key] = values
            for key in array_keys:
                # remove original, unprocessed data from args
                del raw_data[key]
            new_args = [raw_data, *args]
        if 'data' in kwargs:
            # it is also valid to pass data to a form as the data kwarg
            orig_data = kwargs.pop('data') or {}
            array_keys, array_info = self._extract_array_info(orig_data)
            data = orig_data.copy()
            for key, values in array_info.items():
                data[key] = values
            for key in array_keys:
                # remove original, unprocessed data from kwargs
                del data[key]
            new_kwargs['data'] = data
        self.instance = new_kwargs.get('instance')
        super().__init__(*new_args, **new_kwargs)
        self._setup_array_fields()

    def get_silica_config(self):
        if hasattr(self.Meta, 'silica_config'):
            return self.Meta.silica_config
        return None

    def get_field_config(self, field_name):
        silica_config = self.get_silica_config()
        if silica_config:
            return self.Meta.silica_config.get_field_config(field_name)
        return None
    
    def _setup_array_fields(self):
        for field in self.fields.values():
            if isinstance(field, SilicaSubFormArrayField):
                field.parent_instance = self.instance

    def _extract_array_info(self, raw_data):
        if not raw_data:
            return [], {}
        array_keys = []
        # iterate over raw data keys; if any are an array field, then process it
        array_items_by_name_and_count = defaultdict(lambda: defaultdict(dict))
        for key, value in raw_data.items():
            # array fields are named using the following pattern:
            # <array_form_field>.<item_count>.<form_field>
            split_key = key.split('.') # extract pieces
            if len(split_key) < 3:
                # not an array field
                continue
            array_field_name = split_key[0]
            count = split_key[1]
            field = split_key[2]
            array_items_by_name_and_count[array_field_name][count][field] = value
            array_keys.append(key)
        # each array field should have its own list of objects
        array_field_data = {
            array_field_name: values_by_count.values()
            for array_field_name, values_by_count in array_items_by_name_and_count.items()
        }
        return array_keys, array_field_data

    def get_errors_for_template(self):
        return {
            field_name: [e for e in errors] for field_name, errors in self.errors.items()
        }

    def get_data_for_template(self):
        initial = {}
        for field_name, field in self.fields.items():
            if isinstance(field, SilicaSubFormArrayField):
                field.refresh_data()
            # first check instance
            if self.instance and hasattr(self.instance, field_name) and not isinstance(field,
                                                                                       SilicaSubFormArrayField):
                initial[field_name] = getattr(self.instance, field_name)
            # TODO: figure out why this is an empty object for modelforms
            # elif field_name in self.initial:
            #     initial[field_name] = self.initial[field_name]
            else:
                initial[field_name] = field.initial
        return initial

    def get_ui_schema(self):
        # this function is only ever called after the form has been instantiated, so we have access to self.fields
        if hasattr(self.Meta, 'layout'):
            return self.Meta.layout.get_ui_schema(self)
        elements = []
        for field_name, field in self.fields.items():
            ui_kwargs = self._django_widget_to_ui_schema(field, field_config=self.get_field_config(field_name))
            element = Control(field_name, **ui_kwargs)
            elements.append(element)
        return VerticalLayout(*elements).get_ui_schema(self)

    def get_data_schema(self):
        """ Schema is used by the frontend to validate rules """
        # TODO: refine this to support more complex jsonschema rules and to (perhaps) simplify redundant rules
        # TODO: support error_message https://stackoverflow.com/questions/65303161/how-can-i-override-default-error-messages-text-in-json-forms
        properties = {
            field_name: self._django_to_jsonschema_field(field_name, field, field_config=self.get_field_config(field_name))
            for field_name, field in self.fields.items()
        }
        return {
            "type": "object",
            "properties": properties
        }
    
    @property
    def custom_elements(self):
        """ Returns a list of CustomHTMLElement items in the form's Meta.layout """
        if hasattr(self.Meta, 'layout'):
            # we know that the top-level layout cannot be a custom html element, so we can just iterate through elements
            return [e for e in self.Meta.layout.get_all_elements() if isinstance(e, CustomHTMLElement)]
        return []

    def get_custom_elements_content(self):
        custom_elements_content = {}
        for element in self.custom_elements:
            custom_elements_content.update(element.get_mapped_content())
        return custom_elements_content


class SilicaModelFormMixin(SilicaFormMixin, forms.ModelForm):
    def save(self, commit=True):
        super().save(commit=commit)
        # save array fields after we have saved the parent instance
        for field in self.fields.values():
            if isinstance(field, SilicaSubFormArrayField):
                field.do_save()
